/*
  This is a fairly straightforward implementation of a B-Tree in memory modeled after 
  Cormen, et al., Introduction to Algorithms, Chapter 18
 */

#include <stdio.h>

#include <stdlib.h>
#include <stdbool.h>

#include "btree_mem.h"

struct btree_mem_node
   {
   int nkeys;
   void ** key;
   void ** data;
   struct btree_mem_node ** child;
   bool leaf;
   };

struct btree_mem_node * btree_mem_node_new            (struct btree_mem * bt);
void                    btree_mem_node_delete         (struct btree_mem_node * node);

struct btree_mem_node * btree_mem_node_find           (struct btree_mem * bt, struct btree_mem_node * node, void * key, int * i);

void *                  btree_mem_node_insert         (struct btree_mem * bt, struct btree_mem_node * node, void * key, void * data);
void *                  btree_mem_node_insert_nonfull (struct btree_mem * bt, struct btree_mem_node * node, void * key, void * data);

struct btree_mem_node * btree_mem_split_child         (struct btree_mem * bt, struct btree_mem_node * x, int i);

void                    btree_mem_node_walk_inorder   (struct btree_mem_node * node, void (*f)(int, void *, void *), int depth);

////////////////////////////////////////////////////////////////////////////////
struct btree_mem *
btree_mem_new(int min_degree, int (*cmp)(void *, void *))
   {
   struct btree_mem * bt = malloc(sizeof(struct btree_mem));
   if (NULL == bt)
      return NULL;

   bt->mindeg = min_degree;
   bt->cmp = cmp;

   bt->root = btree_mem_node_new(bt);
   if (NULL == bt->root)
      {
      free(bt);
      return NULL;
      }

   bt->root->leaf = true;
   return bt;
   }

////////////////////////////////////////////////////////////////////////////////
// IMPORTANT NOTE: All keys and data must be removed first or else memory
// will leak.
void
btree_mem_delete(struct btree_mem * bt)
   {
   // TODO: walk down the tree and delete all nodes.
   free(bt);
   }

////////////////////////////////////////////////////////////////////////////////
struct btree_mem_node *
btree_mem_node_new(struct btree_mem * bt)
   {
   struct btree_mem_node * node = malloc(sizeof(struct btree_mem_node));
   if (NULL == node)
      return NULL;

   int maxchild = 2*bt->mindeg;

   node->key  = malloc( (maxchild - 1) * sizeof(void *));
   if (NULL == node->key)
      {
      free(node);
      return NULL;
      }

   node->data = malloc( (maxchild - 1) * sizeof(void *));
   if (NULL == node->data)
      {
      free(node->key);
      free(node);
      return NULL;
      }

   node->child = malloc( maxchild * sizeof(struct btree_mem_node *));
   if (NULL == node->data)
      {
      free(node->data);
      free(node->key);
      free(node);
      return NULL;
      }

   for (int i = 0; i < maxchild - 1; i++)
      {
      node->key[i] = NULL;
      node->data[i] = NULL;
      node->child[i] = NULL;
      }
   node->child[maxchild-1] = NULL;

   node->nkeys = 0;
   node->leaf = false;

   return node;
   }

////////////////////////////////////////////////////////////////////////////////
// IMPORTANT NOTE: All keys and data must be removed first or else memory
// will leak.
void
btree_mem_node_delete(struct btree_mem_node * node)
   {
   free(node->child);
   free(node->data);
   free(node->key);
   free(node);
   }

////////////////////////////////////////////////////////////////////////////////
void *
btree_mem_find(struct btree_mem * bt, void * key)
   {
   int i;
   struct btree_mem_node * node = btree_mem_node_find(bt, bt->root, key, &i);
   if (NULL == node)
      return NULL;
   return node->data[i];
   }

////////////////////////////////////////////////////////////////////////////////
struct btree_mem_node *
btree_mem_node_find(struct btree_mem * bt, struct btree_mem_node * node, void * key, int * i)
   {
   *i = 0;
   while (( *i < node->nkeys) && (bt->cmp(key, node->key[*i]) > 0) )
      (*i)++;
   if ( (*i < node->nkeys) && (bt->cmp(key, node->key[*i]) == 0) )
      return node;
   if (node->leaf)
      return NULL;
   else
      return btree_mem_node_find(bt, node->child[*i], key, i);
   }

////////////////////////////////////////////////////////////////////////////////
void *
btree_mem_insert(struct btree_mem * bt, void * key, void * data)
   {
   struct btree_mem_node * s = bt->root;

   printf("inserting >%d< >%s<\n", *((int *) key), (char *) data);
   // If the root node is full, split it here.
   if ((2 * bt->mindeg - 1) == s->nkeys)
      {
      printf("Root node is full: splitting\n");
      s = btree_mem_node_new(bt);
      if (NULL == s)
	 return NULL;
      s->child[0] = bt->root;
      bt->root = s;
      btree_mem_split_child(bt, s, 0);
      }
   return btree_mem_node_insert_nonfull(bt, s, key, data);
   }

////////////////////////////////////////////////////////////////////////////////
void *
btree_mem_node_insert_nonfull(struct btree_mem * bt, struct btree_mem_node * node, void * key, void * data)
   {
   int i = node->nkeys - 1;
   if (node->leaf)
      {
      printf("node is leaf: inserting\n");
      while ( (i >= 0) && (bt->cmp(key, node->key[i]) < 0) )
	 {
	 node->key[i+1] = node->key[i];
	 i--;
	 }
      node->key[i+1] = key;
      node->data[i+1] = data;
      node->nkeys++;
      return key;
      }
   else
      {
      printf("node is not leaf: recursing\n");
      while ( (i >= 0) && (bt->cmp(key, node->key[i]) < 0) )
	 i--;
      i++;
      printf("i %d\n", i);

      /* // Does the child exist?  If not we need to allocate it */
      /* if (NULL == node->child[i]) */
      /* 	 { */
      /* 	 node->child[i] = btree_mem_node_new(bt); */
      /* 	 if (NULL == node->child[i]) */
      /* 	    return NULL; */
      /* 	 } */

      if (node->child[i]->nkeys == (2 * bt->mindeg - 1))
	 {
	 printf("splitting before recursion\n");
	 if (NULL == btree_mem_split_child(bt, node, i))
	    return NULL;
	 if (bt->cmp(key, node->key[i]) > 0)
	    i++;
	 }
      return btree_mem_node_insert_nonfull(bt, node->child[i], key, data);
      }
   }

////////////////////////////////////////////////////////////////////////////////
// x - a non-full node
// i - The index of a full node y in x's child list.
//
// Split y in half, adding the newly created node z as a new child of x.
// z will be placed after y in the child list.

struct btree_mem_node *
btree_mem_split_child(struct btree_mem * bt, struct btree_mem_node * x, int i)
   {

   struct btree_mem_node * y = x->child[i];

   // Allocate new node.
   struct btree_mem_node * z = btree_mem_node_new(bt);
   if (NULL == z)
      return NULL;

   // Load z with the second half of y's data;
   z->leaf = y->leaf;
   z->nkeys = bt->mindeg - 1;
   for (int j = 0; j < bt->mindeg - 1; j++)
      {
      z->key[j] = y->key[j + bt->mindeg];
      z->data[j] = y->data[j + bt->mindeg];
      }
   if (! y->leaf)
      {
      for (int j = 0; j < bt->mindeg; j++)
	 z->child[j] = y->child[j + bt->mindeg];
      }

   // Update y's nkeys.
   y->nkeys = bt->mindeg - 1;

   // Make a hole in x's child list for the new node z.
   for (int j = x->nkeys; j >= i + 1; j--)
      x->child[j+1] = x->child[j];
   x->child[i+1] = z;

   // Make a hole in x's key list for the new node z.
   for (int j = x->nkeys - 1; j >= i; j--)
      {
      x->key[j+1] = x->key[j];
      x->data[j+1] = x->data[j];
      }
   x->key[i] = y->key[bt->mindeg-1];
   x->data[i] = y->data[bt->mindeg-1];

   x->nkeys += 1;

   /* // NULL out the now unused key and data pointers in y. */
   /* for (int j = bt->mindeg; j < (2*bt->mindeg - 1); j++) */
   /*    { */
   /*    y->key[j] = NULL; */
   /*    y->data[j] = NULL; */
   /*    y->child[j] = NULL; */
   /*    } */
   /* y->child[2*bt->mindeg - 1] = NULL; */

   return z;
   }

////////////////////////////////////////////////////////////////////////////////
void *
btree_mem_remove(struct btree_mem * bt, void * key)
   {
   
   }

////////////////////////////////////////////////////////////////////////////////
void
btree_mem_walk_inorder(struct btree_mem * bt, void (*f)(int, void *, void *))
   {
   btree_mem_node_walk_inorder(bt->root, f, 0);
   }

////////////////////////////////////////////////////////////////////////////////
void
btree_mem_node_walk_inorder(struct btree_mem_node * node, void (*f)(int, void *, void *), int depth)
   {

   //   for (int i = 0; i < depth; i++)
   //      printf("\t");
   //   printf("nkeys %d\n", node->nkeys);
   //   printf("walking\n");
   if (0 == node->nkeys)
      {
      //      printf("no keys\n");
      if (NULL != node->child[0])
	 {
	 //	 printf(": walking child\n");
	 btree_mem_node_walk_inorder(node->child[0], f, depth+1);
	 }
      return;
      }

   //   printf("walking %d keys\n", node->nkeys);
   for (int i = 0; i < node->nkeys; i++)
      {
      if (! node->leaf) 
	 btree_mem_node_walk_inorder(node->child[i], f, depth+1);
      f(depth, node->key[i], node->data[i]);
      }
   if (! node->leaf) 
      btree_mem_node_walk_inorder(node->child[node->nkeys], f, depth+1);
   }

